import pybullet
import time
import random
import os
import numpy as np
import pybullet_data
import pybullet_planning

# adapted from pybullet demo: create_visual_shape_array.py

SMALL = 0.1
MEDIUM = 0.5
BIG = 1.

DUCK = "duck"
CUBE = "cube"
WEIGHT = "weight"
BEAR = "bear"

object_dict = {
    "duck": ("duck.obj", "duck_vhacd.obj"),
    "cube": ("cube.obj", "cube.obj"),
    "weight": ("stone.obj", "stone.obj"),
    "bear":("teddy2_VHACD_CHs.obj", "teddy2_VHACD_CHs.obj")
}

def initialize_object(object, size=SMALL):
    try:
        object_tuple = object_dict[object]
    except:
        raise ValueError("""You inputted an invalid object. Make sure you specify a case-sensitive object 
                            from the following list: DUCK, CUBE, WEIGHT, BEAR""")

    pybullet.connect(pybullet.GUI)

    pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
    pybullet.loadURDF("plane_transparent.urdf", useMaximalCoordinates=True)
    #disable rendering during creation.
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)

    shift1 = [0, 0.0, 0]
    shift2 = [0, 0, 0]

    meshScale = [size, size, size]

    # initialize a duck
    visualShapeId = pybullet.createVisualShape(shapeType=pybullet.GEOM_MESH,
                                               fileName=object_tuple[0],
                                               meshScale=meshScale)
    collisionShapeId = pybullet.createCollisionShape(shapeType=pybullet.GEOM_MESH,
                                                     fileName=object_tuple[1],
                                                     meshScale=meshScale)

    assert(visualShapeId == collisionShapeId == 0)

    mb = pybullet.createMultiBody(baseMass=1,
                                  baseInertialFramePosition=[0, 0, 0],
                                  baseCollisionShapeIndex=collisionShapeId,
                                  baseVisualShapeIndex=visualShapeId,
                                  useMaximalCoordinates=False)

    pybullet.changeVisualShape(mb, -1, rgbaColor=[1, 1, 1, 1])
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)
    pybullet.setGravity(0, 0, -10)
    pybullet.setRealTimeSimulation(1)
    return mb

def change_to_colors(colors):
    current_color = 0
    while (1):
        keys = pybullet.getKeyboardEvents()
        if keys and len(colors) > 0:
            key = list(keys.keys())[0]
            if keys[key] == pybullet.KEY_WAS_RELEASED:
                current_color = (current_color + 1) % len(colors)
                pybullet.changeVisualShape(1, -1, rgbaColor=colors[current_color])
                # adapted from pybullet demo: create_visual_shape_array.py

def initialize_environment():
    pybullet.connect(pybullet.GUI)
    pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
    pybullet.loadURDF("plane_transparent.urdf", useMaximalCoordinates=True)
    #disable rendering during creation.
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)
    pybullet.setGravity(0, 0, -10)
    pybullet.setRealTimeSimulation(1)

def spawn_object(obj, size, color):
    try:
        object_tuple = object_dict[obj]
    except:
        raise ValueError("""You inputted an invalid object. Make sure you specify a case-sensitive object 
                            from the following list: DUCK, CUBE, WEIGHT, BEAR""")

    meshScale = [size, size, size]

    # initialize a duck
    visual_id = pybullet.createVisualShape(shapeType=pybullet.GEOM_MESH,
                                            fileName=object_tuple[0],
                                            meshScale=meshScale)
    collision_id = pybullet.createCollisionShape(shapeType=pybullet.GEOM_MESH,
                                                fileName=object_tuple[1],
                                                meshScale=meshScale)

    multibody_id = pybullet.createMultiBody(baseMass=1,
                            baseInertialFramePosition=[0, 0, 0],
                            baseCollisionShapeIndex=collision_id,
                            baseVisualShapeIndex=visual_id,
                            useMaximalCoordinates=False)
    
    pybullet.changeVisualShape(multibody_id, -1, rgbaColor=color)
    return multibody_id, collision_id, visual_id

def any_key_is_pressed():
    keys = pybullet.getKeyboardEvents()
    return any(state & pybullet.KEY_WAS_RELEASED for state in keys.values())

def uniform_size():
    return SMALL

def uniform_object():
    return CUBE

def uniform_color():
    return [1, 1, 1, 1]

def make_objects_of_size_and_color(size_func=uniform_size, object_func=uniform_object, color_func=uniform_color):
    current_color = 0
    while (1):
        if any_key_is_pressed():
            spawn_object(object_func, size_func, color_func)

def launch_box(target_body_id):

    DURATION = 5000

    ALPHA = 100

    # THIS DOES NOT MOVE AT ALL
    # BOX_WIDTH = 0.07
    # BOX_LENGTH = 0.05
    # BOX_HEIGHT = 0.15
    # box_id = pybullet_planning.create_box(BOX_WIDTH, BOX_LENGTH, BOX_HEIGHT, color=pybullet_planning.BLUE)
    # box_geometry = pybullet_planning.get_box_geometry(BOX_WIDTH, BOX_LENGTH, BOX_HEIGHT)
    # collision_id, visual_id = pybullet_planning.create_shape(geometry=box_geometry)
    # box_id = pybullet_planning.create_body(collision_id, visual_id)

    # THIS MOVES, BUT NOT WELL
    box_id, _, _ = spawn_object(uniform_object, uniform_size, uniform_color)

    for i in range(DURATION):
        pybullet.stepSimulation()
        time.sleep(1./240.)
        box_pos, _ = pybullet_planning.get_pose(box_id)
        target_pos, _ = pybullet_planning.get_pose(target_body_id)
        target_pos = np.array(target_pos)
        box_pos = np.array(box_pos)

        force = ALPHA * ((target_pos - box_pos) ** 2)
        # to ensure the force is only applied in the xy direction
        force[2] = 0
        # if i % 100 == 0:
        #     print(force)
        #     print(box_pos, " ", target_pos)
        pybullet.applyExternalForce(objectUniqueId=box_id, linkIndex=-1, forceObj=force, posObj=target_pos, flags=pybullet.LINK_FRAME)
    pybullet.disconnect()
def set_gravity_z_axis(gravitational_constant):
    pybullet.setGravity(0,0,gravitational_constant)

def set_gravity_all_axes(gravity_for_each_axis):
    pybullet.setGravity(*gravity_for_each_axis)

def set_simulation_time_step(time_step):
    pybullet.setTimeStep(time_step)

def simulation_is_running():
    return pybullet.isConnected()
